#!/usr/bin/env perl
# $Header: /home/ed/source_control/psce/bin/diffDist.pl,v 1.2 2005/03/27 20:52:20 ed Exp $

# (c) 2002 Ed Green, Gavin Price UC Regents 
# 2nd part of implementation of 'The bootstrap algorithm for
# estimating standard errors' - Algorithm 6.1 - p.47 of _An
# Introduction to the Bootstrap_, Efron and Tibshirani
# Ed wrote the original covDist.pl code. Gavin adapted it to
# output stats on the difference between two bootstraps rather
# than stats on a single bootstrap and added the confidence 
# interval code.

use Getopt::Std;
use strict;
use Math::Round qw(nearest_ceil nearest_floor);

use vars qw( $opt_e $opt_i $opt_I $opt_j $opt_J $opt_g $opt_b $opt_n $opt_s $opt_h $opt_d);

my $COV_BIN_SIZE = 0.005;
my $DEBUG = 0;
my $Z95 = 1.96;

MAIN : {
    my ( @cov );
    my ( $mu, $std, $CVE_orig);

    &init();
   
    # get array of coverages at specified EPQ level
    ($CVE_orig, @cov) = &readGnuFiles();

    @cov = sort {$a <=> $b} @cov;
    
    if ($DEBUG > 1) { 
	print "Cov Array:\n";
	for (my $i = 0; $i <= $#cov; $i++) {
	    print "$i\t$cov[$i]\n";
	}
    }
    # do some stats
    ( $mu, $std ) = &getStats( @cov );

    my @limits = ($cov[0], $cov[$#cov]);

    my @paramCI = ($mu - $Z95 * $std, $mu + $Z95 * $std);

    my @walkMeanCI = &getWalkCI($mu, \@cov);

    my @walkCVE_CI = &getWalkCI($CVE_orig, \@cov);
    
    my @walkIn_CI = &getWalkIN_CI(\@cov);
    
    # generate output
    &output( $mu, $std, $CVE_orig, \@limits, \@paramCI, \@walkMeanCI, \@walkCVE_CI, \@walkIn_CI, @cov);

}

sub getWalkIN_CI {
    my @cov = @{shift @_};
    my $dist = nearest_floor(1, ($#cov + 1) * 0.05 / 2);
    my $right = $#cov - $dist;
    if ($DEBUG) { print "Walk in L/R index: $dist, $right\n"; }
    return ($cov[$dist], $cov[$right]);
}

sub getWalkCI {
    my $mu = shift;
    my @cov = @{shift @_};
    my ($leftRet, $rightRet);

    my $i = 0;
    while ($cov[$i] < $mu) {
	$i++;
	if( $i > $#cov ) {
	    return (undef, undef);
	}
    }
    if ($DEBUG) { print "Start index: $i\n"; }

    my $dist = nearest_ceil(1, ($#cov + 1) * 0.95 / 2);
    my $left = $i - $dist;
    my $right = $i + $dist - 1;
    
    if ($left < 0) {
	$leftRet = undef;
    }else{
	$leftRet = $cov[$left];
    }
    if ($right > $#cov) {
	$rightRet = undef;
    }else{
	$rightRet = $cov[$right];
    }

    if ($DEBUG) { print "L/R: $left/$right\n";}
    return ($leftRet, $rightRet);
}


sub getStats {
    my ( $sum, $datum, $avg, $foo, $std );
    my @data = @_;


    # find avg
    foreach $datum ( @data ) {
	$sum += $datum;
    }
    $avg = $sum / ($#data + 1);

    # find std
    foreach $datum ( @data ) {
	$foo += (( $datum - $avg ) * ( $datum - $avg )) / $#data;
    }
    $std = sqrt( $foo );
    return ( $avg, $std );
}    

sub output {
    my $mu  = shift;
    my $std = shift;
    my $CVE_orig = shift;
    my $limits = shift;
    my $paramCI = shift;
    my $walkMeanCI = shift;
    my $walkCVE_CI = shift;
    my $walkIn_CI = shift;
    my $e_string = ( "Expected value for coverage difference at $opt_e EPQ is  $mu\n" );
    my $s_string = ( "Standard error for coverage difference at $opt_e EPQ is $std\n" );
    my $c_string = ( "Original value of CVE line at $opt_e EPQ is $CVE_orig\n");
    
    print( "$e_string$s_string$c_string\n" );
    print( "Data min/max: [$limits->[0] , $limits->[1]]\n");
    print( "Parameterized confidence interval: [$paramCI->[0] , $paramCI->[1]]\n");
    print( "CI walking from mean: [$walkMeanCI->[0] , $walkMeanCI->[1]]\n");
    print( "CI walking from CVE line: [$walkCVE_CI->[0] , $walkCVE_CI->[1]]\n");
    print( "CI walking in from max/min data: [$walkIn_CI->[0] , $walkIn_CI->[1]]\n");

    if ( $opt_g ) {
	open( GNUOUT, ">$opt_g.dat" ) or die( "$opt_g.dat: $!" );
	print GNUOUT ( "# $opt_g data in $opt_i at $opt_e\n" );
	print GNUOUT ( "# $e_string\n" );
	print GNUOUT ( "# $s_string\n" );
	my $i;
	my $cov;
	my %cov_bin;
	
	foreach $cov ( @_ ) {
	    $cov_bin{ nearest_floor(1, ( $cov / $COV_BIN_SIZE )) }++;
	    #my $cov_pre = $cov / $COV_BIN_SIZE;
	    #my $cov_temp = nearest_floor(1, ( $cov / $COV_BIN_SIZE ));
	    #print "$cov  \t$cov_pre   \t$cov_temp\t$cov_bin{$cov_temp}\n";
	}
	
	#print "keys";
	my $key;
	my @cov_keys = sort {$a <=> $b} keys %cov_bin;
	for $key (@cov_keys) {  
	    #print "$key ";
	    print  GNUOUT ( $key * $COV_BIN_SIZE); # + 0.5 *$COV_BIN_SIZE );
	    print  GNUOUT ( " " );
	    #print "$cov_bin{$key}\n";
	    unless ( $cov_bin{$key} ) {
		print GNUOUT ( " 0\n" );
	    }
	    else {
		print GNUOUT ( $cov_bin{$key} );
		print GNUOUT ( "\n" );
	    }
	}

	#print GNUOUT ( $key * $COV_BIN_SIZE + 0.5 *$COV_BIN_SIZE );
	#print GNUOUT ( " 0\n" );

	close( GNUOUT );
	
	open ( GS, ">$opt_g.gs" ) or die ( "$opt_g.gs: $!" );
	my $var = $std * $std;
	print GS ( "# Gnuplot script file for data in $opt_i at $opt_e\n" );
	print GS ( "set terminal postscript eps color;\n" );
	print GS ( "set grid;\n" );
	print GS ( "set output \"$opt_g.eps\";\n" );
	print GS ( "plot [0:0.5] \"$opt_g.dat\" with boxes,\\\n" );
	print GS ( "1/($std*sqrt(2*3.14)*exp(((x-$mu)*(x-$mu))/(2*$var))) title \"parameterized\"\n" );
	close ( GS );
    }

}

sub readGnuFiles {
    my $input;
    my @cov;
    
    my $CVE_orig = &getCov("$opt_I/$opt_i$opt_s") 
	- &getCov("$opt_J/$opt_j$opt_s");

    for (my $i = 1; $i <= $opt_b; $i++) {
	if ($DEBUG > 1) { print "$i\n";}
	push( @cov, &getCov( "$opt_I/$opt_i.bbs.$i$opt_s" ) 
	      - &getCov( "$opt_J/$opt_j.bbs.$i$opt_s" ) );
    }
    if ($DEBUG > 1) {print "@cov\n";}
    return ($CVE_orig, @cov);
}

sub getCov {
    my $fn = shift;
    my ( $r, $last_cov );
    my ( @dp );
    my $cur_data_set = 0; #first data set = unnorm, 2nd is lin, 
    # 3rd is quad.  We'll compare this to $opt_n to get the right data
    
    if ( $fn =~ /\.gz?/ ) {
	open( GNU, "gzcat $fn|" ) or die( "$fn: $!" );
    }

    elsif ( $fn =~ /\.bz2?/ ) {
	open( GNU, "bzcat $fn|" ) or die( "$fn: $!" );

    }

    else {
	open( GNU, "$fn" ) or die( "$fn : $!" );
    }
    
    
    while ( $r = <GNU> ) {
	if ( $cur_data_set == $opt_n ) {
	    unless ( $r =~ /\#/ ) {
		@dp = split( /\s+/, $r );
		if ( $dp[ 1 ] >= $opt_e ) {
		    close( GNU );
		    if ($DEBUG > 1) { print "$fn $last_cov\n";}
		    return $last_cov;
		}
		else {
		    $last_cov = $dp[ 0 ];
		}
	    }
	}
	elsif ( $r =~ /\#.+normalized/ ) {
	    $cur_data_set++;
	}
    }
    
    close(GNU);
    if ($DEBUG > 1) { print "$fn $last_cov\n";}
    return $last_cov;
}

sub endsIn {
    my $char = shift;
    my $string = shift;
    my $len = length( $string );
    if ( substr( $string, ($len - 1), 1  ) eq $char ) {
	return 1;
    }
    else {
	return 0;
    }
}

sub init {
    getopts( 'i:j:e:g:b:n:s:h:d:J:I:' );
    unless ( $opt_i && $opt_j && $opt_e && $opt_b) {
	print( "diffDist.pl - Report stats on the coverage distribution\n" );
	print( "at a user specified Errors Per Query (EPQ) rate for the\n" );
	print( "difference between two bootstrap distributions.\n" );
	print( "Base filenames are the file name of the original\n");
	print( "bootstrap files (e.g. filename.bbs.1.dat,\n");
	print( "filename.bbs.2.dat's base filename would be\n");
	print( "'filename'). The filename associated with -j is\n");
	print( "subtracted from the filename associated with -i.\n");
	print( "This code only works on Bayesian bootstrap files.\n");
	print( "OPTIONS:\n" );
	print( "-i [ base filename 1 ]\n" );
	print( "-I [ directory for files with base filename 1; default = ./ ]\n" );
	print( "-j [ base filename 2 ]\n" );
	print( "-J [ directory for files with base filename 2; default = ./ ]\n" );
	print( "-e [ EPQ cutoff ]\n" );
	print( "-g [ filename for gnuplot data file of distribution; if ]\n" );
	print( "     not specified, none is made ]\n" );
	print( "-b [ number of bootstrap repetitions. ]\n" );
	print( "-n [ normalization method; 1 = none, 2 = linear, 3 = quad\n" );
	print( "     default = 2, linear ]\n" );
	print( "-s [ filename suffix after the ID number, e.g. .dat]\n");
	print( "-h [ histogram bin size ]\n");
	exit( 0 );
    }
    
    unless ( $opt_I ) {
	$opt_I = './';
    }
    unless ( $opt_J ) {
	$opt_J = './';
    }

    unless ( $opt_n ) {
	$opt_n = 2;
    }
    if (!$opt_s) { $opt_s = "";}
    if ($opt_d) { $DEBUG = $opt_d;}
    if ($opt_h) { $COV_BIN_SIZE = $opt_h;}
}


