#!/usr/bin/env perl
# $Header: /home/ed/source_control/psce/bin/maxCov.pl,v 1.2 2005/03/27 20:52:20 ed Exp $

use strict;
use Getopt::Std;
use vars qw( $opt_i $opt_e $opt_r );

MAIN : {

    my ( %un_norm, %lin_norm, %quad_norm );
    my ( @gnufiles );
    my ( $gnufile, $found_un, $found_lin, $found_quad, 
	 $r, $cov, $epq, $last_cov, $comp );

    &init();

    opendir( GNU, $opt_i ) or die( "$opt_i: $!" );
    @gnufiles = readdir( GNU );
    closedir( GNU );

    foreach $gnufile ( @gnufiles ) {
        if ( &hasRest( $gnufile ) ) {
	    if ( $comp = &isCompressed( "$opt_i$gnufile" ) ) {
		open( GNU, "$comp"."cat $opt_i$gnufile|" ) or 
		    die( "$comp"."cat $opt_i$gnufile: $!" );
	    }
	    else {
		open( GNU, "$opt_i$gnufile" ) or 
		    die( "$opt_i$gnufile: $!" );
	    }
	    
	    ( $found_un, $found_lin, $found_quad ) = ( 0, 0, 0 );

	    # Go to unnormalized section
	    until ( $found_un ) {
		$r = <GNU>;
		unless ( $r =~ /^\#/ ) {
		    $last_cov = $cov;
		    ( $cov, $epq ) = split( /\s+/, $r );
		    if ( $epq > $opt_e ) {
			$un_norm{ $gnufile } = $last_cov;
			$found_un = 1;
		    }
		}
	    }
	    
	    # Go to linearly normalized section
	    $r = <GNU>;
	    until ( $r =~ /^\# Linearly normalized/ ) {
		$r = <GNU>;
	    }
	    until ( $found_lin ) {
		$r = <GNU>;
		$last_cov = $cov;
		( $cov, $epq ) = split( /\s+/, $r );
		if ( $epq > $opt_e ) {
		    $lin_norm{ $gnufile } = $last_cov;
		    $found_lin = 1;
		}
	    }

	    # Go to quadratically normalized section
	    $r = <GNU>;
	    until ( $r =~ /^\# Quadratically normalized/ ) {
		$r = <GNU>;
	    }
	    
	    until ( $found_quad ) {
		$r = <GNU>;
		$last_cov = $cov;
		( $cov, $epq ) = split( /\s+/, $r );
		if ( $epq > $opt_e ) {
		    $quad_norm{ $gnufile } = $last_cov;
		    $found_quad = 1;
		}
	    }

	    close( GNU );
				
	}
    }

    print( "Unnormalized:\n" );
    foreach $gnufile ( sort {$un_norm{$b} <=> $un_norm{$a} } keys %un_norm ) {
	print( "$gnufile => ", $un_norm{ $gnufile }, "\n" );
    }

    print( "Linearly normalized:\n" );
    foreach $gnufile ( sort {$lin_norm{$b} <=> $lin_norm{$a} } keys %lin_norm ) {
	print( "$gnufile => ", $lin_norm{ $gnufile }, "\n" );
    }

    print( "Quadratically normalized:\n" );
    foreach $gnufile ( sort {$quad_norm{$b} <=> $quad_norm{$a} } keys %quad_norm ) {
	print( "$gnufile => ", $quad_norm{ $gnufile }, "\n" );
    }

}

sub hasRest {
    my $gnufile = shift;
    if ( index( $gnufile, $opt_r ) != -1 ) {
	return 1;
    }
    else {
	return 0;
    }
}

sub endsIn {
    my $char = shift;
    my $string = shift;
    my $len = length( $string );
    if ( substr( $string, ($len - 1), 1  ) eq $char ) {
	return 1;
    }
    else {
	return 0;
    }
}

# determines if an input file is compressed by checking its suffix
# returns the root of the cat program that will cat this compressed
# file if it is compressed.  Returns false if it is not compressed
sub isCompressed {
    my $fn = shift;
    if ( substr( $fn, length( $fn ) - 2 ) eq 'gz' ) {
	return 'gz';
    }
    elsif ( substr( $fn, length( $fn ) - 3 ) eq 'bz2' ) {
	return 'bz';
    }
    else {
	return 0;
    }
}

sub init {
    getopts( 'i:e:r:' );
    unless ( $opt_i ) {
	print( "maxCov.pl\n" );
	print( "Shows the coverage(s) at a user-specified error rate for\n" );
	print( "all the CVE data in a user-specified directory.\n" );
	print( "OPTIONS:\n" );
	print( "-i [ Input directory with gnuplot datafiles in\n" );
	print( "     EPQ_maker.pl format.\n" );
	print( "-e [ EPQ level to examine; default = 0.01 ]\n" );
	print( "-r [ filename restriction for data files ]\n" );
	exit( 0 );
    }
    unless ( &endsIn( '/', $opt_i ) ) {
	$opt_i .= '/';
    }

    unless ( $opt_e ) {
	$opt_e = 0.01;
    }
}
    
