#!/usr/bin/env perl
# $Header: /home/ed/source_control/psce/bin/splitAstralSFs.pl,v 1.2 2005/03/27 20:52:20 ed Exp $
# (c) 2004 Ed Green / UC Regents

use strict;
use Getopt::Std;
use vars qw( $opt_d );

MAIN : {
    &init();

    my @cur;      # buffer of lines of current fasta sequence entry
    my $line;
    my @cur_scop; # scop info for last 
    my @scop;     # scop info for this entry
    my @head;     # things on the header line
    my $a = 1;    # BOOLEAN => TRUE IF WE SHOULD SEND
                  # current buffer to .a
    my $sa = 1;   # BOOLEAN => TRUE IF WE SHOULD SEND
                  # current buffer to .a
    my $fh;
    my $DEBUG = 1;

    open( DBIN, $opt_d ) or die ( "$opt_d: $!" );
    open( DBOUTA, ">$opt_d.a" ) or die( "$opt_d.a: $!" );
    open( DBOUTB, ">$opt_d.b" ) or die( "$opt_d.b: $!" );
    $fh = \*DBOUTA;

    while ( <DBIN> ) {
	# next entry - deal with the last one
	if ( /^>/ ) {
	    @head = split( /\s+/ );
	    @scop = split( /\./, $head[ 1 ] );
	    # If this entry is in the same SF as the last one, then
	    # send it to whatever $a says and toggle $a
	    if ( &sameSF( \@scop, \@cur_scop ) ) {
		if ( $a ) {
		    &addEntryToDb( $fh, @cur );
		    $fh = \*DBOUTA;
		    $a = 0;
		}
		else {
		    &addEntryToDb( $fh, @cur );
		    $fh = \*DBOUTB;
		    $a = 1;
		}
	    }
	    # This is a new SF
	    else {
		if ( $sa ) {
		    &addEntryToDb( $fh, @cur );
		    print ( "a" ) if $DEBUG;
		    $fh = \*DBOUTA;
		    $a = 0;
		    $sa = 0;
		}
		else {
		    &addEntryToDb( $fh, @cur );
		    print ( "b" ) if $DEBUG;
		    $fh = \*DBOUTB;
		    $a = 1;
		    $sa = 1;
		}
	    }
	    @cur = ();
	    @cur_scop = @scop;
	}
	push( @cur, $_ );
    }

    # Now deal with this last one
    &addEntryToDb( $fh, @cur );

    close ( DBIN );
    close ( DBOUTA );
    close ( DBOUTB );

}

# Takes a filehandle and an array of lines that comprises a sinlge,
# complete fasta entry, header and all.  Writes the lines to the
# filehandle.  Returns nothing.
sub addEntryToDb {
    my $fh = shift;
    foreach my $l ( @_ ) {
	print $fh ( $l );
    }
}

# Takes pointer to two arrays: the SCOP classifications of the
# sequence we've just seen and that of the last sequence.  Returns
# true if they're the same in the class, fold, and superfamily
# fields.  Returns false otherwise.
sub sameSF {
    my $scop_p   = shift;
    my $c_scop_p = shift;

    # If we haven't seen a current scop yet (first one)
    return 1 unless $c_scop_p->[ 0 ];

    # Check for sameness of class, fold, and SF fields
    # Compare as string since some of these fields are letters, some
    # are numbers
    if ( ( $scop_p->[ 0 ] eq $c_scop_p->[ 0 ] ) &&
	 ( $scop_p->[ 1 ] eq $c_scop_p->[ 1 ] ) &&
	 ( $scop_p->[ 2 ] eq $c_scop_p->[ 2 ] ) ) {
	return 1;
    }
    else {
	return 0;
    }
}

sub init {
    getopts( 'd:' );
    unless( $opt_d ) {
	print( "splitAstralSFs.pl\n" );
	print( "Takes a fasta formatted ASTRAL database\n" );
	print( "with SCOP classification info on the\n" );
	print( "header line.  Creates two new databases\n" );
	print( "each having approximately half the entries\n" );
	print( "of the source.\n" );
	print( "Each superfamily is split in half.  Each output\n" );
	print( "database has half the members of each superfamily.\n" );
	print( "Single member superfamilies are alternately sent\n" );
	print( "to each output database.\n" );
	print( "The output databases that are created have the same\n" );
	print( "name as the original with the suffixes .a and .b\n" );
	print( "OPTIONS:\n" );
	print( "-d [ input database - REQUIRED ]\n" );
	exit( 0 );
    }
}
